use image::{ImageBuffer, Rgb, RgbImage};

trait Distance<const N: usize> {
    fn dist(&self, a: [f64; N], b: [f64; N]) -> Result<f64, &'static str>;
}

struct HilbertPoly {
    vertices: Vec<[f64; 2]>
}

#[allow(dead_code)]
struct Euclidean<const N: usize>();

impl<const N: usize> Distance<N> for Euclidean<N> {
    fn dist(&self, a: [f64; N], b: [f64; N]) -> Result<f64, &'static str> {
        Ok(a.iter().zip(b.iter()).map(|x| (x.1 - x.0) * (x.1 - x.0)).sum::<f64>().sqrt())
    }
}

#[allow(dead_code)]
struct Manhattan<const N: usize>();

impl<const N: usize> Distance<N> for Manhattan<N> {
    fn dist(&self, a: [f64; N], b: [f64; N]) -> Result<f64, &'static str> {
        Ok(a.iter().zip(b.iter()).map(|x| f64::abs(x.1 - x.0)).sum::<f64>())
    }
}

impl<const N: usize> Distance<N> for HilbertPoly {
    fn dist(&self, a: [f64; N], b: [f64; N]) -> Result<f64, &'static str> {

        // 2d Hilbert Convex Polygon, 3d is slightly different as it will be to a side (edge of
        // convex set )
        if N != 2 {
            return Err("Dim != 2");
        };

        // Get edges.
        let mut edges = self.vertices.windows(2).map(|x| [x[0], x[1]]).collect::<Vec<[[f64; 2]; 2]>>();
        edges.push([self.vertices[0], self.vertices[self.vertices.len() - 1]]);

        // Assert that the Hilbert convex set is convex.
        let mut prev = 0.0;
        let mut curr;
        for i in 0..N {
            let edge_one = edges[i];
            let edge_two = edges[(i + 1) % N];
            curr = (edge_one[1][0] - edge_one[0][0]) * (edge_two[1][1] - edge_one[0][1]) - 
                (edge_one[1][1] - edge_one[0][1]) * (edge_two[1][0] - edge_one[0][0]);
            if curr != 0.0 {
                if curr * prev < 0.0 {
                    return Err("Not convex.");
                } else  {
                    prev = curr;
                }
            }
        }

        // Check if points are both inside the set
        let (mut i, mut j, mut c) = (0, self.vertices.len() - 1, false);
        loop {
            if i >= self.vertices.len() {
                break;
            }


            if ((self.vertices[i][1] > a[1]) != (self.vertices[j][1] > a[1])) && (a[0] < (self.vertices[j][0] - self.vertices[i][0]) * (a[1] - self.vertices[i][1]) / (self.vertices[j][1] - self.vertices[i][1]) + self.vertices[i][0]) {
                c = !c;
            }
            j = i;
            i += 1;
        }

        if !c {
            return Err("First point not in set.");
        }


        let (mut i, mut j, mut c) = (0, self.vertices.len() - 1, false);
        loop {
            if i >= self.vertices.len() {
                break;
            }


            if ((self.vertices[i][1] > b[1]) != (self.vertices[j][1] > b[1])) && (b[0] < (self.vertices[j][0] - self.vertices[i][0]) * (b[1] - self.vertices[i][1]) / (self.vertices[j][1] - self.vertices[i][1]) + self.vertices[i][0]) {
                c = !c;
            }
            j = i;
            i += 1;
        }

        if !c {
            return Err("Second point not in set.");
        }

        // Intersect the line with each polygon side
        let slope = (b[1] - a[1]) / (b[0] - a[0]);
        let p_intersect = edges.into_iter().map(|p| {
            let edge_slope = (p[1][1] - p[0][1]) / (p[1][0] - p[0][0]);
            let d = edge_slope - slope;
            if d == 0.0 {
                None
            } else {
                Some([((edge_slope * p[0][0] - p[0][1]) - (slope * a[0] - a[1])) / d, ((slope * (edge_slope * p[0][0] - p[0][1])) - (edge_slope * (slope * a[0] - a[1]))) / d])
            }
        }).filter(|x| x.is_some()).map(|x| x.map_or([0.0, 0.0], |x| x)).collect::<Vec<[f64; 2]>>();

        // Find the closest distance
        let a_closest = p_intersect.clone().into_iter().map(|p| (p, (p[0] - a[0]) * (p[0] - a[0]) + (p[1] - a[1]) * (p[1] - a[1]))).min_by(|a, b| a.1.partial_cmp(&b.1).unwrap()).unwrap();
        let b_closest = p_intersect.into_iter().map(|p| (p, (p[0] - b[0]) * (p[0] - b[0]) + (p[1] - b[1]) * (p[1] - b[1]))).min_by(|a, b| a.1.partial_cmp(&b.1).unwrap()).unwrap();

        let pb = f64::sqrt((a_closest.0[0] - b[0]) * (a_closest.0[0] - b[0]) + (a_closest.0[1] - b[1]) * (a_closest.0[1] - b[1]));
        let pa = f64::sqrt(a_closest.1);
        let aq = f64::sqrt((b_closest.0[0] - a[0]) * (b_closest.0[0] - a[0]) + (b_closest.0[1] - a[1]) * (b_closest.0[1] - a[1]));
        let bq = f64::sqrt(b_closest.1);

        Ok(f64::ln((pb * aq) / (pa * bq)))
    }
}

fn naive_voronoi_2d(metric: &mut dyn Distance<2>, w: u32, h: u32, points: Vec<(u32, u32)>) {
    let mut image: RgbImage = ImageBuffer::new(w, h);
    // All points must be in the width / height
    for p in points {
        assert!(p.0 <= w);
        assert!(p.1 <= h);
    }

    for i in 0..w {
        for j in 0..h {

        }
    }
}

fn main() {

    let triangle = HilbertPoly { vertices: vec![[-8.0, 0.0], [8.0, 0.0], [0.0, 8.0]] };
    println!("{}", triangle.dist([-2.0, 2.0], [2.0, 2.0]).unwrap());
}
